<?php
///V:1.0.0
///M:REA Parser
///H:reaParser.html
///S:$REA_USE_PARSER

define('kParser_Operators', "=:!><().,+-/\^*&%{}[];");

//Token types
define('kParser_EOF', 255);
define('kParser_EOL', 13);
define('kParser_String', 1);
define('kParser_Operator', 2);
define('kParser_Comment', 3);
define('kParser_Number', 4);
define('kParser_Identifier', 5);
define('kParser_Keyword', 6);
define('kParser_Function', 6);

class ckBasicParser {
	var $source='';
	var $done = false;
	var $abort = false;
	var $_debug = false;
	var $keywords = array('-----');
	var $saved_token = null;
	var $last_token = null;

	function syntaxError($msg){
		$this->abort = true;
		print $msg . "\n";
	}
		
	function isWhite($c){
		if(($c == "\n") || ($c == "\r") || ($c == "\t") || ($c == ' ')) return true;
		return false;
	}
	function getCh($p){
		$c = $this->source[$p];
		//print " &nbsp; {$p}-" . ord($c) . "-[$c]<br>";
		return $c;
	}
	function isKeyword($k){
		if (in_array(strtolower($k), $this->keywords)) return true;
		return false;
	}
	function needToken($tok){
		$tk = $this->nextToken();
		if($tk === false) return false;
		
		if($tk['value'] != $tok){
			$this->putTokenBack();
			$this->syntaxError("Expecting token \"{$tok}\" but found \"{$tk['value']}\". " . $tok);
			return false;
		}
		
		return true;
	}
	function ifOperator($op){
		$tk = $this->nextToken();
		if($tk === false) return false;
		
		if($tk['type'] != kParser_Operator){
			$this->putTokenBack();
			return false;
		}
		if($tk['value'] != $op){
			$this->putTokenBack();
			return false;
		}
		
		return true;
		
	}
	function skipNextToken(){
		$tk = $this->nextToken();
	}
	function pokeToken(){
		$tk = $this->nextToken();
		$this->putTokenBack();
		
		return $tk;
	}
	function putTokenBack(){
		if(empty($this->last_token)) return;
		$this->saved_token = $this->last_token;
		//print "saving {$this->saved_token['value']}\n";
	}
	function nextToken(){
		$this->last_token = null;
		if(!empty($this->saved_token)){
			//print "returning old {$this->saved_token['value']}<br>";
			$tk = $this->saved_token;
			$this->saved_token = null;
			$this->last_token = $tk;
			return $tk;
		}
		$tk = array('type'=> kParser_EOF, 'value'=>'');
		$c = ''; $b = ''; $p = -1;
		$sl = strlen($this->source) - 1;
		if($sl < 0){
			//print "at eof<br>";
			return false;
		}

		do{
			$p++;
			//$c = $this->source[$p];
			$c = $this->getCh($p);
			
		}while($this->isWhite($c) && ($p <= $sl));
		
		
		if($p > $sl){
			$this->done = true;
			$this->source = "";
			return false;
		}

		if($c == chr(34)){
			$tk['type'] = kParser_String;
			$b = $c;
			do{
				$p++;
				//$c = $this->source[$p];
				$c = $this->getCh($p);
				//print " &nbsp; {$p}-" . ord($c) . "-[$c]<br>";
				$b.= $c;
			}while(($c != chr(34))  && ($p <= $sl));

			$this->source = substr($this->source, $p+1);
			$tk['value'] = $b;
		}elseif($c == chr(39)){ //comment
			//print "GOT Comment, source[{$p}] = $c<br>";
			$tk['type'] = kParser_Comment;
			do{
				$p++;
				$c = $this->source[$p];
			}while( (($c != "\n") && ($c != "\r")) && ($p <= $sl) );
			
			//print "End Comment, source[{$p}] = $c<br>";
			$tk['value'] = '';
			$this->source = substr($this->source,$p+1);
		}elseif(strpos(kParser_Operators, $c) !== false){
			if($c == '<'){
				if($this->source[$p+1] == '>'){
					$c = '!=';
					$p++;
				}elseif($this->source[$p+1] == '='){
					$c = '<=';
					$p++;
				}
			}elseif($c == '>'){
				if($this->source[$p+1] == '='){
					$c = '>=';
					$p++;
				}
			}elseif($c == '!'){
				if($this->source[$p+1] == '='){
					$c = '!=';
					$p++;
				}
			}elseif($c == ':'){
				if($this->source[$p+1] == ':'){
					$c = '::';
					$p++;
				}
			}elseif($c == '='){
				if($this->source[$p+1] == '='){
					$c = '==';
					$p++;
				}
				if($this->source[$p+1] == '>'){
					$c = '=>';
					$p++;
				}
			}elseif($c == '+'){
				if($this->source[$p+1] == '='){
					$c = '+=';
					$p++;
				}
			}elseif($c == '-'){
				if($this->source[$p+1] == '='){
					$c = '-=';
					$p++;
				}elseif($this->source[$p+1] == '>'){
					$c = '->';
					$p++;
				}
			}elseif($c == '*'){
				if($this->source[$p+1] == '='){
					$c = '*=';
					$p++;
				}
			}elseif($c == '/'){
				if($this->source[$p+1] == '='){
					$c = '/=';
					$p++;
				}
			}elseif($c == '&'){
				if($this->source[$p+1] == '&'){
					$c = '&&';
					$p++;
				}
			}elseif($c == '|'){
				if($this->source[$p+1] == '|'){
					$c = '||';
					$p++;
				}
			}elseif($c == ';'){
				
			}
			$tk['type'] = kParser_Operator;
			$tk['value'] = $c;
			$this->source = substr($this->source, $p+1);
		}elseif($c == '$'){
			$b = '';
			while(!$this->isWhite($c) && !(strpos(kParser_Operators, $c) !== false) && !($c==' ') && !($c==';')){
				$b .= $c;
				if($p > $sl) break;
				$p++;
				//$c = $this->source[$p];
				$c = $this->getCh($p);
			}
			
			$tk['type'] = kParser_Identifier;
			$tk['value'] = $b;
			
			$this->source = substr($this->source, $p);
		}else{
			$b = '';
			while(!$this->isWhite($c) && !(@strpos(kParser_Operators, $c) !== false) && !($c==' ')  && !($c==';')){
				$b .= $c;
				if($p > $sl) break;
				$p++;
				//$c = $this->source[$p];
				$c = $this->getCh($p);
			}
			
			$this->source = substr($this->source, $p);
			if(strpos('0123456789.', $b[0]) !== false){
				$tk['type'] = kParser_Number;
				$tk['value'] = $b;
			}elseif($b[0] == '@'){
				$tk['type'] = kParser_String;
				$tk['value'] = substr($b,1);
			}elseif((strtolower($b) == 'and') || (strtolower($b) == 'or')){
				$tk['type'] = kParser_Operator;
				$tk['value'] = strtolower($b);
			}elseif($this->isKeyword($b)){
				$tk['type'] = kParser_Keyword;
				$tk['value'] = strtolower($b);
			}elseif(strtolower($b) == 'true'){
				$tk['type'] = kParser_Number;
				$tk['value'] = '1';
			}elseif(strtolower($b) == 'false'){
				$tk['type'] = kParser_Number;
				$tk['value'] = '0';
			}else{
				$tk['type'] = kParser_Identifier;
				$tk['value'] = $b;
			}
		}
		$this->last_token = $tk;
		return $tk;
	}

}

?>