<?php
require_once('ckparser.php');
require_once('ckglobal.php');
require_once('ckruntime.php');

class ckbasic extends ckBasicParser {
	var $symtab = array();
	var $src_block = '';
	var $src_dim = '';
	var $curr_class = false;
	function setup(){
		global $rt_lang_keywords;
		$this->keywords = $rt_lang_keywords;
	}
	function source(){
		
		return $this->src_dim . $this->src_block;
	}
	function convert($src){
		$this->source = $src;
		//$tk  =  $this->nextToken();
		while($this->nextToken() !== false){
			if($this->abort) break;
			$this->statement();
		}
		
	}
	function statement(){
		$tk = $this->last_token;
		if( $this->last_token['type'] == kParser_Identifier ){
			$this->convAssigment();
		}elseif( $this->last_token['type'] == kParser_Keyword ){
			$this->convStatement($this->last_token);
			
		}elseif( $this->last_token['type'] == kParser_Operator ){
			$this->src_block .= $this->last_token['value'];
		}elseif( $this->last_token['type'] == kParser_Comment ){
			
		}else{
			$this->src_block .= $this->last_token['value'] . ';';
		}
	}

	function convStatement($tk){
		$type = $tk['type'];
		$s = $tk['value'];
		
		if( ($s == 'dim') || ($s == 'function') || ($s == 'method') || ($s == 'with')){
			$this->convDefine($s);
			return;
		}elseif( ($s == 'model') ){
			$this->convClass();
			return;
		}elseif( ($s == 'if') || ($s == 'elseif') || ($s == 'else') ){
			$this->convIF($s);
			return;
		}elseif( ($s == 'print') || ($s == 'msgbox') || ($s == 'echo') ){
			$this->convPrint();
			return;
		}elseif( ($s == 'return') ){
			$this->convReturn();
			return;
		}elseif($s == 'for'){
			$this->convFor();
			return;
		}elseif($s == 'foreach'){
			$this->convForEach();
			return;
		
		}elseif($s == 'end'){
			$this->convEnd();
			return;
		}elseif($s == 'while'){
			$this->convWhile();
			return;
		}elseif( ($s == 'break') || ($s == 'continue') ){
			$this->src_block .= $s . ';'  . "\n";
			return;
		}elseif( ($s == 'exit') ){
			$this->src_block .= 'exit;'  . "\n";
			return;
		}elseif($type == kParser_Identifier){
			$this->convAssigment($tk);
		}
		
	}
	function convAssigment(){
		$tk = $this->last_token;
		$tok = $tk['value'];
		$sn = $this->parseVarSymbol($tok);
		$is_obj = ($sn != $tok) ? true : false;
		
		$m = false;
		if( $this->ifOperator('(') ){
			$sn = $this->parseFunctionArguments($sn,'(');
			$m = true;
		}
		if( $this->ifOperator('[') ){
			$sn = $this->parseFunctionArguments($sn,'[');
		}
		
		while( $this->ifOperator('.') ){
			$tk = $this->pokeToken();
			if(strtolower($tk['value']) == 'count'){
				$sn = 'count(' . $sn . ')';
				$this->skipNextToken();
			}else{
				//$sn = $this->parseVarSymbol('($_m=' . $sn .')', false);
				$sn = $this->parseVarSymbol($sn, false);
			}
		}
				
		$tk = $this->pokeToken();
		if($tk['type'] == kParser_Operator){
			$this->skipNextToken();
			
			$op = $tk['value'];
			$sv = $this->parseRValue();
			$this->src_block .= $sn . ' ' . $op . ' ' . $sv . ';' . "\n";
			return;
		}
		
		
		if(!$is_obj && $m) $sn = substr($sn,0);
		$this->src_block .= $sn . ';'  . "\n";
		//$this->src_block .= $sn;
		return;
		
		
		
	}
	function convDefine($type){
		if($type == 'dim'){
			$this->convDim();
			return;
		}elseif($type == 'with'){
			$this->convDim();
			return;
		}elseif($type == 'function'){
			$this->convFunction();
			return;
		}elseif($type == 'method'){
			$this->convFunction();
			return;
		}
		
		
	}
	function convEnd(){
		$tk = $this->nextToken();
		if($tk['type'] != kParser_Keyword){
			$this->syntaxError('Expecting keyword for END statement');
			return;
		}
		
		switch($tk['value']){
			case 'sub':
			case 'function':
			case 'message':
			case 'method':
			case 'event':
				$this->src_block.= '}' . "\n";
				break;
			case 'model':
				$this->curr_class = false;
				$this->src_block.= '}' . "\n";
				break;
		}	
	}
	function convDim(){
		do{
			$v = $this->parseDim(true);
			$this->src_block.= $v . ';' . "\n";
						
			if( !$this->ifOperator(',') ) break;
		}while($this->last_token !== false);
	}
	function convWith(){
		do{
			$v = $this->parseDim(true);
			$this->src_block.= $v . ';' . "\n";
						
			if( !$this->ifOperator(',') ) break;
		}while($this->last_token !== false);
	}	
	
	function convArray(){
		$out = 'array(';
		if(!$this->needToken('(') ){
			return '';
		}
		
		do{
			$v = $this->parseRValue(true);
			$out.= $v;
			
			$tk = $this->pokeToken();
			if($tk['value'] == '=>'){
				$this->skipNextToken();
				$out.= ' => ' . $this->parseRValue(true);
			}
			if( !$this->ifOperator(',') ) break;
			
			$out.= ",";
			
		}while($this->last_token !== false);
		
		if(!$this->needToken(')') ){
			return $out . ')';
		}
		
		$out.=")";
		
		return $out;
		
	}
	function convPrint($tr = true){
		$sv = $this->parseRValue();
		$this->src_block.= 'print ' . $sv;
		if($tr) $this->src_block.= ';' . "\n";
	}
	
	function convReturn($tr = true){
		$sv = $this->parseRValue();
		$this->src_block.= 'return ' . $sv;
		if($tr) $this->src_block.= ';' . "\n";
	}
	
	function convClass(){
		$tk = $this->nextToken();
		if($tk['type'] != kParser_Identifier){
			$this->syntaxError('Expecting Identifier for class name');
			return;
		}
		$sn = $tk['value'];
		
		$st = 'object';
		
		$tk = $this->pokeToken();		
		if($tk['value'] == 'extends'){
			$this->skipNextToken();
			$tk = $this->nextToken();
			if($tk['type'] != kParser_Identifier){
				$this->syntaxError('Expecting Identifier for class name');
				return;
			}
			$st = $tk['value'];
			if($st[0] == '$') $st = substr($st,1);
		}
		$sn1 = substr($sn, 1);
		$this->src_dim .= "{$sn} = rt_nc('{$sn1}', '{$st}');\n";
		
		
		//$this->src_block.= '$' . $sn . '->= new ' . $st . ";\n";

	}
	function convFunction(){
		$tk = $this->nextToken();
		if($tk['type'] != kParser_Identifier){
			$this->syntaxError('Expecting Identifier for function name1');
			return;
		}
		
		$scls = false;
		$sn = $tk['value'];
		
		$tk = $this->pokeToken();
		if($tk['value'] == '::'){
			$this->skipNextToken();
			$scls = $sn;
			$tk = $this->nextToken();
			if($tk['type'] != kParser_Identifier){
				$this->syntaxError('Expecting Identifier for function name2');
				return;
			}
			$sn = $tk['value'];
		}
		
		if(!$this->needToken('(') ){
			return;
		}
		$this->func_params = array('$me = $rt_params[0];', '$super = $rt_params[1];');
		
		do{
			$tk = $this->nextToken();
			if($tk === false) break;
			if( $tk['value'] == ')' ){
				$this->putTokenBack();
				break;
			}else{
				$sv = $this->parseFormalParameter(false);	
			}
			
			$spar .= $sv;

			$tk = $this->nextToken();
			if($tk === false) break;
			if( $tk['value'] != ',' ){
				$this->putTokenBack();
				break;
			}else{
				$spar.= ', ';
			}
			
			if($this->abort) break;
		}while(true);
		
		if(!$this->needToken(')')) return;
		//if(!$this->needToken('{')) return;
		if($scls !== false){
			
			$this->src_dim.= '$' . $scls . "->object_selectors[] = '{$sn}';\n";
			$sn = 'rt_' . $scls . '_' . $sn;
			
			$spar = implode("\n", $this->func_params);
			$this->src_block.= 'function ' . $sn . '($rt_params=array()){' . "\n" . $spar . "\n";
		}else{
			$this->src_block.= 'function ' . $sn . '(' . $spar . '){' . "\n";
		}
		
		$this->func_params = array();
		
		
	}
	function convWhile(){
		
		$s = 'while(';
		$s.= $this->parseRValue();
		$s.= '){';
		$this->src_block.= $s  . "\n";
	
		do{
			$tk = $this->nextToken();
			if($tk === false) break;
			if( $tk['value'] == 'wend' ){
				$this->src_block.= "}\n";
				break;
			}
			
			$this->statement();
			if($this->abort) break;
		}while(true);
		
	}
	function convForEach(){
		
		$s = 'foreach(';
		
		$tk = $this->nextToken();
		if($tk['type'] != kParser_Identifier){
			$this->syntaxError('Expecting Identifier ForEach loop');
			return;
		}
		$sc1 = $tk['value'];
		$sc1 = $this->parseVarSymbol($sc1);
		
		$tk = $this->pokeToken();
		if($tk['value'] == '=>'){
			$this->skipNextToken();
			
			$tk = $this->nextToken();
			if($tk !== false){
				if($tk['type'] != kParser_Identifier){
					$this->syntaxError('Expecting Identifier ForEach loop');
					return;
				}
				$sc2 = $tk['value'];
				$sc2 = $this->parseVarSymbol($sc2);
				
				$sc1 = "$sc1 => $sc2";
			}
		}
		
		if(!$this->needToken('in')) return;
		$s1= $this->parseRValue();
		
		$s.= $s1 . ' as ' . $sc1;
		$s.= '){';
		$this->src_block.= $s  . "\n";
	
		do{
			$tk = $this->nextToken();
			if($tk === false) break;
			if( $tk['value'] == 'next' ){
				$this->src_block.= "}\n";
				break;
			}
			
			$this->statement();
			if($this->abort) break;
		}while(true);
		
	}
	function convFor(){
		
		$s = 'for(';
		
		$tk = $this->nextToken();
		if($tk['type'] != kParser_Identifier){
			$this->syntaxError('Expecting Identifier as control variable in FOR loop');
			return;
		}
		$sc = $tk['value'];
		$sc = $this->parseVarSymbol($sc);
		
		if(!$this->needToken('=')) return;
		$sv = $this->parseRValue();
		
		$s.= $sc . ' = ' . $sv . '; ';
		
		$op_tk = $this->nextToken();
		if($op_tk['value'] == 'to'){
			$op = ' += ';
			$cm = ' <= ';
		}elseif($op_tk['value'] == 'downto'){
			$op = ' -= ';
			$cm = ' >= ';
		}else{
			$this->syntaxError('Expecting "To" or "DownTo" operator');
			return;
		}
		
		$sto = $this->parseRValue();
		
		$tk = $this->pokeToken();
		if($tk['value'] == 'step'){
			$this->skipNextToken();
			$si = $this->parseRValue();
		}else{
			$si = '1';
		}
			
		$s.= $sc . $cm . $sto . '; ';
		$s.= $sc . $op . $si . '){';
		$this->src_block.= $s  . "\n";
		
		do{
			$tk = $this->nextToken();
			if($tk === false) break;
			if( $tk['value'] == 'next' ){
				$this->src_block.= "}\n";
				break;
			}
			
			$this->statement();
			if($this->abort) break;
		}while(true);
		
	}
	function convIF($st){
		if($st != 'else'){
			$sc = $this->parseRValue();
			$tk = $this->pokeToken();
			if($tk['value'] != 'then'){
				$this->src_block.= $st . ' (' . $sc . ');' . "\n";
				return;
			}
			$this->skipNextToken();
			$s = $st . ' (' . $sc . '){';
			if($st == 'elseif') $s = '}' . $s;
		}else{
			$s = '}else{';
		}
		$this->src_block.= $s . "\n";
		do{
			$tk = $this->nextToken();
			if($tk === false) break;
			if( ($tk['value'] == 'else') || ($tk['value'] == 'elseif') ){
				$this->putTokenBack();
				break;
			}
			if( $tk['value'] == 'end' ){
				$tk = $this->pokeToken();
				if($tk['value'] == 'if'){
					$this->skipNextToken();
					$this->src_block.= "}\n";
					break;
				}
			}
			
			$this->statement();
		}while(true);
		
	}
	function parseFunction($s){
		//only appear on the right side
		if( ($s == 'print') || ($s == 'msgbox') || ($s == 'echo') ){
			
		}elseif( ($s == 'break') || ($s == 'continue') ){
			return $s;
			
		}elseif( ($s == 'false') || ($s == 'true') ){
			return $s;
		}elseif( ($s == 'exit') ){
			return 'exit';
		}elseif( ($s == 'array') ){
			return $this->convArray();
		}
	}	
	function parseFormalParameter($et = false){
		if($et){
			$tk = $this->nextToken();
		}else{
			$tk = $this->last_token;
		}
		
		$sname = $tk['value'];
		
		$i = count($this->func_params);
		
		$tk = $this->nextToken();
		if($tk['value'] == 'as'){
			$tk = $this->nextToken();
		}elseif($tk['value'] == '='){
			$sv = $this->parseRValue();
			
			$init = "if(count(\$rt_params) <= $i){\n{$sname} = {$sv};\n}else{\n{$sname}=\$rt_params[$i];\n}";
			$this->func_params[] = $init;
			
			$sname .= ' = ' . $sv;
			
		}else{
			$this->putTokenBack();
			$this->func_params[] = "{$sname}=\$rt_params[$i];";
		}
		
		
		return $sname;
	}

	function parseDim($et = false){
		if($et){
			$tk = $this->nextToken();
		}else{
			$tk = $this->last_token;
		}
		
		$sn = $tk['value'];
		$sname = $sn;
		$this->symtab[$sname] = array();
		
		$tk = $this->nextToken();
		if($tk['value'] == 'as'){
			$tk = $this->nextToken();
			if($tk['value'] == 'object'){
				$this->symtab[$sname] = array('type'=>'object');
				return "{$sname} = rt_no('{$sn}')";
			}elseif(($tk['value'] == 'integer') || ($tk['value'] == 'boolean') || ($tk['value'] == 'long') || ($tk['value'] == 'float') ){
				return "{$sname} = 0";
			}elseif( ($tk['value'] == 'string') || ($tk['value'] == 'text') ){
				return "{$sname} = ''";
			}else{
				$this->symtab[$sname] = array('type'=> $tk['value']);
				return "{$sname} = rt_ni('{$sn}', '{$tk['value']}')";
			}
		}else{
			$this->putTokenBack();
		}
		return "{$sname} = rt_no('{$sn}')";
	}
	function parsePHPSymbol($sObject, $pd = true){
		
		$tk  =  $this->nextToken();
		$smember = $tk['value'];
		
		if( !$this->ifOperator('(') ){
			//if(substr($sObject,0,1) == chr(34)) return $sObject;
			//if(strpos('0123456789.', $sObject[0]) !== false) return $sObject;
			return $sObject . "->{$smember}";
		}
		
		$sv = $sObject . "->{$smember}(";
		if( $this->ifOperator(')') ){
			$sv.= ')';
			return $sv;
		}
		
		while($this->last_token !== false){
			if( $this->ifOperator(')') ) break;
			
			if( $this->ifOperator(',') ){
				$sv.= ', ';
			}else{
				$v = $this->parseRValue();
				$sv.= $v;
			}
		}
		$sv.= ')'; 
		return $sv;
	}
	function parseVarSymbol($sObject, $pd = true){
		if($pd){
			if( !$this->ifOperator('.') ){
				
				if( $this->ifOperator('->') ) return $this->parsePHPSymbol($sObject, $pd);
				
				if(substr($sObject,0,1) == chr(34)) return $sObject;
				if(strpos('0123456789.', $sObject[0]) !== false) return $sObject;
				return $sObject;
			}
		}
		$tk  =  $this->nextToken();
		$smember = $tk['value'];
		
		if( !$this->ifOperator('(') ){
			//if(substr($sObject,0,1) == chr(34)) return $sObject;
			//if(strpos('0123456789.', $sObject[0]) !== false) return $sObject;
			return $sObject . "->{$smember}";
		}
		
		$sv = "rt_sm({$sObject}, \"{$smember}\"";
		if( $this->ifOperator(')') ){
			$sv.= ')';
			return $sv;
		}else{
			$sv.= ', ';
		}
		
		while($this->last_token !== false){
			if( $this->ifOperator(')') ) break;
			
			if( $this->ifOperator(',') ){
				$sv.= ', ';
			}else{
				$v = $this->parseRValue();
				$sv.= $v;
			}
		}
		$sv.= ')'; 
		return $sv;
	}
	function parseRValue(){
		return $this->parseExpression();
		
	}
	function parseFunctionArguments($sf, $del='('){
		$sf .= $del;
		$ed = ($del == '(') ? ')' : ']';
		
		$tk  =  $this->last_token;
		while($this->last_token !== false){
			if( $this->ifOperator($ed) ) break; 
			if( $this->ifOperator(',') ){
				$sf.= ',';
			}else{
				$sv = $this->parseRValue();
				$sf.= $sv;
			}
			if($this->abort) break;
			//$tk  =  $this->nextToken();
		}
		
		$sf.= $ed;
			
		if( $this->ifOperator('.') ){
			return $this->parseVarSymbol($sf, false);
		}
		return $sf;
		
	}

	
	function parseExpression(){

		
		$tk  =  $this->nextToken();

		$neg = '';

		if($tk['type'] == kParser_Operator){
			if( ($tk['value'] == '+') || ($tk['value'] == '-') ){
				$neg = $tk['value'];
			}elseif($tk['value'] == '!'){
				$neg = $tk['value'];
			}else{
				$this->putTokenBack();
			}
		}else{
			$this->putTokenBack();
		}

		$v1 = $this->parseTerm();
		if($neg != '') $v1 = $neg . $v1;
		
		$tk  =  $this->nextToken();
		
	
		
		$op_ex = array('and', 'or', '==', '=', '!=', '>=', '<=', '>', '<', '+', '-', '&', '&&');
		if( in_array($tk['value'], $op_ex) ){
			//noting
		}else{
			$this->putTokenBack();
			if($this->_debug) print "<b>End</b> parseExpression</blockquote>";
			return $v1;
		}
		
		while( in_array($tk['value'], $op_ex) ){
			if($this->_debug) print "At parseExpression, with operator: {$tk['value']}<br>";
			
			$v2 = $this->parseTerm();
			
			if($tk['value'] == 'and'){
				$v1 = $v1 . ' && ' . $v2;
			}elseif($tk['value'] == 'or'){
				$v1 = $v1 . ' || ' . $v2;
			}elseif($tk['value'] == '='){
				$v1 = $v1 . ' == ' . $v2;
			}elseif($tk['value'] == '=='){
				$v1 = $v1 . ' == ' . $v2;
			}elseif( ($tk['value'] == '!=') || ($tk['value'] == '<>') ){
				$v1 = $v1 . ' != ' . $v2;
			}elseif($tk['value'] == '&'){
				$v1 = $v1 . ' . ' . $v2;
			}else{
				$v1 = $v1 . ' ' . $tk['value'] . ' ' . $v2;
			}

			$tk  =  $this->nextToken();
			if($this->_debug) print "At parseExpression, next operator: {$tk['value']}<br>";
			//print "while [{$tk['value']}]<br>";
		}
		
		$this->putTokenBack();
		return $v1;
	}// end parseExpression
	function parseTerm(){
		$v1 = $this->parseFactor();
		$tk  =  $this->nextToken();
		while( in_array($tk['value'], array('*', '/', '^', '%') )){
			$v1 = $v1 . ' ' . $tk['value'] . ' ' . $this->parseFactor();
			$tk  =  $this->nextToken();
		}
		$this->putTokenBack();
		return $v1;
	}
	function parseFactor(){

		if($this->_debug) print "<b>At parseFactor</b><blockquote>";
		$tk  =  $this->nextToken();
		$type = $tk['type'];
		$tok = $tk['value'];
		
		if($tok == '('){
			$v1 = '(' . $this->parseExpression() . ')';
			$this->needToken(')');
			
		}elseif($tok == '['){
			$v1 = '[' . $this->parseExpression() . ']';
			$this->needToken(']');
		}elseif($type == kParser_Number){
			$v1 = $tok;
		}elseif($type == kParser_String){
			$v1 = $tok;
		}elseif($type == kParser_Identifier){
			$v1 = $tok;
			
			$v1 = $this->parseVarSymbol($v1);
			$is_obj = ($v1 != $tok) ? true : false;
			$is_fn = false;
			if($this->ifOperator('(') ){
				$v1 = $this->parseFunctionArguments($v1);
				$is_fn = true;
			}elseif($this->ifOperator('[') ){
				//print "\nhello array\n";
				$v1 = $v1 . '[' . $this->parseExpression() . ']';
				$this->needToken(']');
				
				while( $this->ifOperator('.') ){
					$tk = $this->pokeToken();
					if(strtolower($tk['value']) == 'count'){
						$v1 = 'count(' . $v1 . ')';
						$this->skipNextToken();
					}else{
						$v1 = $this->parseVarSymbol($v1, false);
					}
				}
			}elseif($this->ifOperator('::') ){
				
				do{
					$tk = $this->pokeToken();
					print "***** Poked " . strtolower($tk['value']) . "\n";
					if(strtolower($tk['value']) == 'count'){
						$v1 = 'count(' . $v1 . ')';
						$this->skipNextToken();
					}else{
						$v1 = $this->parseVarSymbol($v1, false);
					}
				}while( $this->ifOperator('.') );
				
			}elseif($this->ifOperator('.') ){
				
				do{
					//$v1 = $this->parseVarSymbol('($_m=' . $v1 .')', false);
					$v1 = $this->parseVarSymbol($v1, false);
				}while( $this->ifOperator('.') );
			}
			if(!$is_obj && $is_fn){
				//$v1 = substr($v1, 1);
			}
			
		}elseif( $type == kParser_Keyword ){
			$v1 = $this->parseFunction($tok);
		}else{
			$this->putTokenBack();
			return 0;
		}
		return $v1;
	}//end parseFactor
	
	function &getIdentifier($tok){
		if( array_key_exists($tok, $this->symtab) ) return $this->symtab['1'];
		
	}
	function handleMember($tok){
		
	}
}
?>